"""
A file watching service to rename movie files and move them
to relevant locations after match the file against the porndb.
"""

import os
import re
import shutil
import sys
import tempfile
import time
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import List, Optional

import schedule
from loguru import logger
from watchdog.events import EVENT_TYPE_DELETED, EVENT_TYPE_MOVED, FileSystemEvent, PatternMatchingEventHandler
from watchdog.observers.polling import PollingObserver

from namer.fileutils import is_interesting_movie, make_command_relative_to, move_command_files
from namer.namer import process_file
from namer.types import Command, default_config, NamerConfig
from namer.web.server import WebServer


def done_copying(file: Optional[Path]) -> bool:
    """
    Determines if a file is being copied by checking its size in 2 second
    increments and seeing if the size has stayed the same.
    """
    if file is None or file.exists() is False:
        return False
    while True:
        try:
            # pylint: disable=consider-using-with
            buffered_reader = open(file, mode="rb")
            buffered_reader.close()
            break
        except PermissionError:
            time.sleep(0.2)
    return True


@logger.catch
def handle(command: Command):
    """
    Responsible for processing and moving new movie files.
    """
    process_file(command)


def retry_failed(namer_config: NamerConfig):
    """
    Moves the contents from the failed dir to the watch dir to attempt reprocessing.
    """
    logger.info("Retry failed items:")
    # remove all old namer log files
    for log_file in namer_config.failed_dir.rglob("**/*_namer.log"):
        log_file.unlink()
    # move all files back to watch dir.
    for file in list(namer_config.failed_dir.iterdir()):
        shutil.move(namer_config.failed_dir / file.name, namer_config.watch_dir / file.name)


def is_fs_case_sensitive():
    """
    Create a temporary file to determine if a filesystem is case-sensitive, or not.
    """
    with tempfile.NamedTemporaryFile(prefix="TmP") as tmp_file:
        return not Path(tmp_file.name.lower()).is_file()


class MovieEventHandler(PatternMatchingEventHandler):
    """
    When a new movie file is detected, this class handles the event,
    and passes off the processing to the handler() function above.
    """

    namer_config: NamerConfig

    def __init__(self, namer_config: NamerConfig, queue: Queue):
        super().__init__(patterns=["*.*"], case_sensitive=is_fs_case_sensitive(), ignore_directories=True, ignore_patterns=None)
        self.namer_config = namer_config
        self.command_queue = queue

    def on_any_event(self, event: FileSystemEvent):
        file_path = None
        if event.event_type == EVENT_TYPE_MOVED:
            file_path = event.dest_path  # type: ignore
        elif event.event_type != EVENT_TYPE_DELETED:
            file_path = event.src_path
        if file_path is not None:
            path = Path(file_path)
            relative_path = str(path.relative_to(self.namer_config.watch_dir))
            if re.search(self.namer_config.ignored_dir_regex, relative_path) is None and done_copying(path) and is_interesting_movie(path, self.namer_config):
                logger.info("watchdog process called for {}", relative_path)
                # Extra wait time in case other files are copies in as well.
                if self.namer_config.del_other_files is True:
                    time.sleep(self.namer_config.extra_sleep_time)
                self.prepare_file_for_processing(path)

    @logger.catch
    def prepare_file_for_processing(self, path: Path):
        command = make_command_relative_to(input_dir=path, relative_to=self.namer_config.watch_dir, config=self.namer_config)
        working_command = move_command_files(command, self.namer_config.work_dir)
        if working_command is not None:
            self.command_queue.put(working_command)


class MovieWatcher:
    """
    Watches a configured dir for new files and attempts
    to process them and move to a destination dir.

    If a failure occurs moves the new files to a failed dir.

    See NamerConfig
    """

    def __processing_thread(self):
        while True:
            command = self.__command_queue.get()
            if command is None:
                break
            handle(command)
            self.__command_queue.task_done()
        self.__command_queue.task_done()

    def __init__(self, namer_config: NamerConfig):
        self.__started = False
        self.__stopped = False
        self.__namer_config = namer_config
        self.__src_path = namer_config.watch_dir
        self.__event_observer = PollingObserver()
        self.__webserver: Optional[WebServer] = None
        self.__command_queue: Queue = Queue()
        self.__worker_thread: Thread = Thread(target=self.__processing_thread, daemon=True)
        self.__event_handler = MovieEventHandler(namer_config, self.__command_queue)
        self.__background_thread: Optional[Thread] = None

    def run(self):
        """
        Checks for new files in 3 second intervals,
        needed if running in docker as events aren't properly passed in.
        """
        if not self.__started:
            self.start()
            if self.__namer_config.web is True:
                self.__webserver = WebServer(self.__namer_config, self.__command_queue)
                self.__webserver.start()
            try:
                while True and not self.__stopped:
                    schedule.run_pending()
                    time.sleep(3)
                self.stop()
            except KeyboardInterrupt:
                self.stop()
            self.__started = False
            self.__stopped = False

    def __enter__(self):
        self.__background_thread = Thread(target=self.run)
        self.__background_thread.start()
        tries = 0
        while self.get_web_port() is None and tries < 20:
            time.sleep(0.2)
            tries += 1
        if self.get_web_port is None:
            raise RuntimeError("application did not get assigned a port within 4 seconds.")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        if self.__background_thread is not None:
            self.__background_thread.join()
            self.__background_thread = None

    def start(self):
        """
        starts a background thread to check for files.
        """
        config = self.__namer_config
        logger.info("Start porndb scene watcher.... watching: {}", config.watch_dir)
        if os.environ.get("PROJECT_VERSION"):
            project_version = os.environ.get("PROJECT_VERSION")
            print(f"Namer version: {project_version}")
        if os.environ.get("BUILD_DATE"):
            build_date = os.environ.get("BUILD_DATE")
            print(f"Built on: {build_date}")
        if os.environ.get("GIT_HASH"):
            git_hash = os.environ.get("GIT_HASH")
            print(f"Git Hash: {git_hash}")
        self.__schedule()
        self.__event_observer.start()
        self.__worker_thread.start()
        # touch all existing movie files.
        files: List[Path] = []
        for file in self.__namer_config.watch_dir.rglob("**/*.*"):
            if file.is_file() and file.suffix.lower()[1:] in self.__namer_config.target_extensions:
                files.append(file)
        for file in files:
            if file.exists() and file.is_file():
                self.__event_handler.prepare_file_for_processing(file)

    def stop(self):
        """
        stops a background thread to check for files.
        Waits for the event processor to complete.
        """
        if not self.__stopped:
            self.__stopped = True
            logger.info("exiting")
            self.__event_observer.stop()
            self.__event_observer.join()
            if self.__webserver is not None:
                self.__webserver.stop()
            self.__command_queue.put(None)
            self.__command_queue.join()
            logger.info("exited")

    def get_web_port(self) -> Optional[int]:
        if self.__webserver is not None:
            return self.__webserver.get_effective_port()

    def __schedule(self):
        self.__event_observer.schedule(self.__event_handler, str(self.__src_path), recursive=True)


def create_watcher(namer_watchdog_config: NamerConfig) -> MovieWatcher:
    """
    Configure and start a watchdog looking for new Movies.
    """
    logger.remove()
    logger.add(sys.stdout, format="{time} {level} {message}", level="INFO")
    logger.info(str(namer_watchdog_config))
    if not namer_watchdog_config.verify_watchdog_config():
        sys.exit(-1)
    if namer_watchdog_config.retry_time is not None:
        schedule.every().day.at(namer_watchdog_config.retry_time).do(lambda: retry_failed(namer_watchdog_config))
    movie_watcher = MovieWatcher(namer_watchdog_config)
    return movie_watcher


if __name__ == "__main__":
    create_watcher(default_config()).run()



"""
A wrapper allowing shutdown of a Flask server.
"""
import mimetypes
from queue import Queue
from threading import Thread
from typing import Optional, Union

from flask import Flask
from flask_compress import Compress
from waitress import create_server
from waitress.server import BaseWSGIServer, MultiSocketServer

from namer.types import NamerConfig
from namer.web.routes import get_web_routes

app = Flask(__name__)
compress = Compress()


class WebServer:
    """
    A wrapper allowing shutdown of a Flask server.
    """
    __server: Union[MultiSocketServer, BaseWSGIServer]
    __thread: Thread
    __config: NamerConfig
    __command_queue: Queue

    __mime_types: dict = {
        'font/woff': '.woff',
        'font/woff2': '.woff2',
    }

    def __init__(self, config: NamerConfig, command_queue: Queue):
        self.__config = config
        self.__command_queue = command_queue
        self.__add_mime_types()
        self.__make_server()

    def __make_server(self):
        path = '/' if self.__config.web_root is None else self.__config.web_root
        blueprint = get_web_routes(self.__config, self.__command_queue)
        app.register_blueprint(blueprint, url_prefix=path, root_path=path)
        compress.init_app(app)
        self.__server = create_server(app, host=self.__config.host, port=self.__config.port)
        self.__thread = Thread(target=self.__run, daemon=True)

    def __add_mime_types(self):
        for mime, ext in self.__mime_types.items():
            test_mime, test_ext = mimetypes.guess_type(f'0{ext}')
            if test_mime is None:
                mimetypes.add_type(mime, ext)

    def start(self):
        self.__thread.start()

    def __run(self):
        """
        Start server on existing thread.
        """
        if self.__server:
            self.__server.run()

    def stop(self):
        """
        Stop severing requests and empty threads.
        """
        if self.__server:
            self.__server.close()

    def get_effective_port(self) -> Optional[int]:
        return getattr(self.__server, "effective_port", None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['namer', 'namer.web']

package_data = \
{'': ['*'],
 'namer.web': ['public/*',
               'public/assets/*',
               'public/assets/fonts/*',
               'templates/components/*',
               'templates/pages/*',
               'templates/partials/*',
               'templates/render/*']}

install_requires = \
['Flask-Compress>=1.12,<2.0',
 'Pillow>=9.1.1,<10.0.0',
 'Unidecode>=1.3.4,<2.0.0',
 'flask>=2.1.2,<3.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'lxml>=4.8.0,<5.0.0',
 'mutagen>=1.45.1,<2.0.0',
 'pathvalidate>=2.5.0,<3.0.0',
 'rapidfuzz>=2.0.5,<3.0.0',
 'requests-cache>=0.9.4,<0.10.0',
 'requests>=2.27.1,<3.0.0',
 'schedule>=1.1.0,<2.0.0',
 'waitress>=2.1.1,<3.0.0',
 'watchdog>=2.1.6,<3.0.0']

setup_kwargs = {
    'name': 'namer',
    'version': '1.5.10',
    'description': 'A namer of video files based on metadata from the porndb.',
    'long_description': '.. |logo| image:: ./logo/namer.png\n  :width: 80\n  :class: display: inline flow; align: left\n\n##############\n|logo| Namer\n##############\n\n.. image:: https://github.com/ThePornDatabase/namer/actions/workflows/ci.yml/badge.svg?\n  :target: https://github.com/ThePornDatabase/namer/actions/workflows/ci.yml/\n.. image:: https://codecov.io/gh/ThePornDatabase/namer/branch/main/graph/badge.svg?token=4MQEN2NUKZ\n  :target: https://codecov.io/gh/ThePornDatabase/namer\n.. image:: https://badge.fury.io/py/namer.svg?\n  :target: https://badge.fury.io/py/namer\n.. image:: https://img.shields.io/pypi/dm/namer?logo=pypi&logoColor=fff\n  :target: https://pypi.org/project/namer\n.. image:: https://static.pepy.tech/personalized-badge/namer?period=total&units=international_system&left_color=grey&right_color=yellowgreen&left_text=Downloads\n  :target: https://pepy.tech/project/namer  \n.. image:: https://static.pepy.tech/personalized-badge/namer?period=month&units=international_system&left_color=grey&right_color=yellowgreen&left_text=Downloads/Month\n  :target: https://pepy.tech/project/namer\n\nNamer is a powerful command line tool for renaming and tagging mp4 video files in a way that helps plex/jellyfin/emby and related plugins extract that data or lookup data with the PornDB_\'s plugins for plex or jellyfin/emby.\nNamer is easily installed as a python pip and can:\n\n* can be used to name and embed tags in individual files with metadata from porndb:\n\n  ``python -m namer rename -f /path/to/file/Site.[YY]YY.MM.DD.MessOfText.XXX.2160.mp4 [-v]`` \n\n* can be used to name and tag files with metadata from a jellyfin/emby/kodi .nfo file (should be named the same as the file except for extension).\n\n  ``python -m namer rename -f /path/to/file/Site.[YY]YY.MM.DD.MessOfText.XXX.2160.mp4 -i [-v]``\n\n* can be used to rename a tag a file based on the directory name, so if you have a file like ``/Site.[YY]YY.MM.DD.MessOfText.XXX.2160/abc.mp4`` \n\n  ``python -m namer rename -d /path/to/dir/Site.[YY]YY.MM.DD.MessOfText.XXX.2160/``\n\n* can be used to rename a tag a whole mess of dirs and files in a directory (using -m, meaning "many").\n\n  ``python -m namer rename -m -d /path/to/dir/``\n  \n* can be used to just suggest a possible name.  The file doesn\'t need to exist but should have an extension.\n\n  ``python -m namer suggest -f Site.[YY]YY.MM.DD.MessOfText.XXX.2160.mp4``\n\n* can be run watching a directory for new files to name, tag and move to an output location, possible setting file permissions, writing .nfo files with downloaded images, attempting to grab trailers, and retrying failed files nightly.\n\n  ``python -m namer watchdog``  \n\nFor all of the above it\'s recommended to have a config file in your home directory (copied from namer.cfg.sample in this git repo)\n\nAlso provided is a docker file if you prefer.\n\nIt is possible to ignore and not need to parse dates for studios added to a list in the configuration file.  This is mostly used for studios that do not list dates on videos.\n\nWhy should I use this?\n----------------------\n\n1.  You have partially well structured file names (say from an rss feed, etc) and you never want to have to manually match files in plex/jellyfin/emby with the PornDB_\'s plugin.\n2.  You don\'t want your recent videos to be added to your library until they are matchable in the PornDB_.\n3.  You want to store the metadata about a file in mp4 files, in a way that can be read by Apple TV app, including information like: Studio, date created, name, performers, original url, proper HD tags, ratings, and movie poster.   All of this data is readable by Plex, and most by Jellyfin/Emby in case you want to standard Apple video players or your library\'s metadata storage is ever damaged.\n\nHow successful at matching videos is this tool?\n------------------------------------------------\n\nFor data pulled from the internet with rss feeds (which are often in the file format listed below) .... perfect.  The author and others have only experienced one mismatch, and that type of failure can never occur again.\n\nIf running in a background watchdog mode, files that were failed to match are retried every 24 hours, letting the PornDB_ scrapers catch up with any metadata they may be missing.\n\nOptionally, a log file can be enabled to show the original file name parts, what options were evaluated, and which match was used to name the file, it will be written next to your video file with the same name as the file (with a _namer.log) suffix rather than an mp4/mkv/avi/mov/flv extension.   This is very useful for sanity checking matches, and if ever a mismatch does occur the original file name is available in the log.\n\n\nFor the curious, how is a match made?\n------------------------------------------------\n\nIt assumes that file names exist as in a format like ```sitename-[YY]YY-MM-DD-Scene.and.or.performer.name.mp4.```.  A powerful regex tries to determine the various parts of a file\'s name.   Note that the separating dashes and dots above are interchangeable, and spaces may also be used as separators (or any number of any combo of the three.)   This regex is overridable, but you really need to know what you\'re doing and if you don\'t have all the match groups for the regex, the match from the the PornDB_ will likely not be any where near as robust as it is with a site, a date, and a scene/perform name section.\nYou\'ll have to read the code to figure out how to set this.   You really shouldn\'t do it.\n\nWhen determining a possible queried match from the PornDB_:\n\nSitename my not need be the full name of the site, as long as a the looked up sitename starts with file\'s sitename it could be a valid match.\n\nThe date may have a four digit or two digit year.  If two digit, "20" is assumed as the default century, not "19".  A potential match must be with one day plus/minus the file\'s date to be considered a match.\n\nFinally the looked up scene name and all performers first and last names are combined in to what is called a powerset (every combo of including or not including each artist and/or scene name), and that is compared against the file\'s \'Scene.and.or.performer.name\' section with a tool called rapidfuzz.   A name must be 90% similar to a member of the powerset to be considered a match, though all potential matches are evaluated and sorted before selecting the best match.   Information about all potential matches are stored in the local log file if it is enabled.\n\nI\'m sold how do I install it!\n--------------------------------------------------\n\nYou have two choices.   Do you use docker?  Pull the docker image, here\'s docker-compose snippet to help you out.\n\n\n.. code-block:: yaml\n\n  version: "3"\n  services:  \n    namer:\n      container_name: namer\n      image: ghcr.io/theporndatabase/namer:latest\n      environment:\n        - PUID=1001\n        - PGID=1000\n        - TZ=America/Los_Angeles\n        - NAMER_CONFIG=/config/namer.cfg\n      volumes:\n        - /apps/namer/:/config <- this will store the namer.cfg file.\n        - /media:/data <- this will have the four folders namer needs to work, referenced in the config file.\n      restart: always\n\nCopy namer.cfg to your config location (a path mapped to /config/namer.cfg above), and set values for your setup.   \nThe config is well commented and you should only need to add a token for the porndb and change file locations.\n\nRunning a service will occur automatically once you call ``docker-compose up``.  Now check out the configuration section below.\n\n\nPip/Python usage\n--------------------\n\nWhat if you don\'t want to use docker and/or containers?  Do you have python 3 and pip (sometimes pip3) and the command line tool ``ffmpeg`` installed locally?  If so,  ``pip install namer`` gets the job done.  If\nyou don\'t have python (3), pip and ``ffmpeg`` installed Homebrew_ can help you on Mac, and Chocolatey_ can help you on windows\n\n\n.. code-block:: sh\n\n  # install namer\n  pip3 install namer\n\n  #optionallly, set your configuration location, the below is the default:\n  export NAMER_CONFIG=${HOME}/.namer.cfg  \n\n  # Run the watchdog service:\n  python3 -m namer watchdog\n\n  # Or manually rename a file, dir, or all sub-dirs/sub-files of a dir:\n  # This calls the help method so that you can see the options.\n  python3 -m namer rename -h\n\n\nConfiguration:\n---------------------------\n\nThere is a well documented template of namer.cfg in this git repo, which is broken up in to three sections.\nOne section is related to command line renaming, the `namer section`_, one related to tagging mp4s `metadata section`_, \nand finally one related to the watchdog process `watchdog section`_.\nPlease note that the `namer section`_ section and the `watchdog section`_ \nsection both have a field to describe the new name of a file based on looked up metadata from the PornDB_.   \nThey differ because when run from the command line namer will keep the file "in place".  \nIf namer is passed a dir on the command line as input it can operate in one of two modes,\nthe default mode is to look for the largest mp4 file, or other configured movie file extension if no mp4 exists,\nand rename and move that file to the root of the folder (if it\'s in a sub-folder).\nIn this case, by default the assumption is the name of the folder should be parsed to look for information to\nsearch the PornDB_ for matching rather than the file name.   Meaning,\nif you pass a file to namer on the commandline it will be renamed but stay in the same directory.\n\n\nTypical Watchdog Behavior:\n----------------------------\n\nThe watchdog process will watch a single folder, configured with watch_dir_ in the ``namer.cfg`` file.   Any new files and directories that appear in the watch_dir_\nwill be processed once an mp4/mkv/avi/mov/flv file has been fully copied in to it.  \n\nThe first step in processing is to more the newly appearing directory or file in to the work_dir_.\n\nOnce moved the processing is highly dependant on the namer.cfg file, but in general, the name of video file or the directory file (configured with ``prefer_dir_name_if_available`` flag)\nis parsed and matched with a scene from the PornDB_.   See `For the curious, how is a match made?`_.  If a match cannot be made the general assumption is that the PornDB_ doesn\'t have metadata for that file yet.\nThe file is move to the failed dir fail_dir_ to be retried once a day at a time configured with retry_time_,\nwhich by default will be a random selected minute in the 3am hour of your timezone.   If enabled_tagging_ flag is set to true then\nthe metadata (including cover art if enable_poster_ is set) will be embedded in the mp4 file.  Please read the comments in the namer.cfg to find out about genres, tags, performers, etc.\n\nFinally, the file is moved to a location defined by dest_dir_ and new_relative_path_name_.\n\n\nDevelopment\n------------------------------\n\n.. code-block:: sh\n\n  # Building:\n  poetry build\n\n  # Linting:\n  poetry run flakeheaven lint\n\n  # Testing:\n  poetry run pytest\n\n  # Code Coverage:\n  poetry run pytest --cov\n\n  # Formatting:\n  poetry run autopep8 --in-place namer/*.py test/*.py\n\n  # Html Coverage report:\n  poetry run coverage html\n\n  # Publishing:\n  # First make sure you have set gotten a token from pypi and set it on your machine.\n  poetry config pypi-token.pypi <token>\n\n  # Perhaps update the version number?\n\n  # Publishing a release to pypi.org:\n  poetry publish\n\n  # build docker file with:\n  ./docker_build.sh\n\nPull Requests Are Welcome!\n---------------------------\n\nJust be sure to pay attention to the tests and any failing pylint results.   If you want to vet a pr will be accepted before building code, file an new feature request issue, and 4c0d3r will comment on it and set you up for success.   Tests are must.\n\n.. _PornDB: http://metadataapi.net/\n.. _namer section: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L1\n.. _metadata section: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L59\n.. _watchdog section: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L89\n.. _watch_dir: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L100\n.. _work_dir: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L104\n.. _fail_dir: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L109\n.. _dest_dir: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L112\n.. _retry_time: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L115\n.. _new_relative_path_name: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L97\n.. _enabled_tagging: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L67\n.. _enable_poster: https://github.com/ThePornDatabase/namer/blob/main/namer.cfg.sample#L72\n.. _Homebrew: https://docs.brew.sh/Installation\n.. _Chocolatey: https://chocolatey.org/install\n',
    'author': '4c0d3r',
    'author_email': '4c0d3r@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
ffmpeg is access through this file, it is used to find the video stream's resolution,
and update audio streams "Default" setting.   Apple video players require there be
only one default audio stream, and this script lets you set it with the correct language
code there are more than one audio streams and if they are correctly labeled.
See:  https://iso639-3.sil.org/code_tables/639/data/
"""

import json
import shutil
import string
import subprocess
from pathlib import Path
from random import choices
from types import SimpleNamespace
from typing import Optional

from loguru import logger


def get_resolution(file: Path) -> int:
    """
    Gets the vertical resolution of a mp4 file.  For example, 720, 1080, 2160...
    Returns zero if resolution can not be determined.
    """
    logger.info("resolution stream of file {}", file)

    with subprocess.Popen(
        [
            "ffprobe",
            "-v",
            "error",
            "-select_streams",
            "v:0",
            "-show_entries",
            "stream=height",
            "-of",
            "csv=p=0",
            file,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    ) as process:
        stdout, stderr = process.communicate()
        success = process.returncode == 0
        output = 0
        if stdout is not None:
            output = stdout
        if not success:
            logger.warning("Error getting resolution of file {}", file)
            if stderr is not None:
                logger.warning(stderr)
        if output is not None:
            logger.info("output {}", output)
            return int(output)
    return 0


def get_audio_stream_for_lang(mp4_file: Path, language: str) -> int:
    """
    given a mp4 input file and a desired language will return the stream position of that language in the mp4.
    if the language is None, or the stream is not found, or the desired stream is the only default stream, None is returned.
    See: https://iso639-3.sil.org/code_tables/639/data/

    Returns -1 if stream can not be determined
    """

    with subprocess.Popen(
        [
            "ffprobe",
            "-show_streams",
            "-select_streams",
            "a",
            "-of",
            "json",
            "-i",
            mp4_file,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    ) as process:
        stdout, stderr = process.communicate()
        success = process.returncode == 0
        audio_streams_str = None
        if stdout is not None:
            audio_streams_str = stdout
        if not success:
            logger.warning("Error getting audio streams of file {}", mp4_file)
            if stderr is not None:
                logger.warning(stderr)
        logger.info("Target for audio: {}", mp4_file)
        audio_streams = None
        if audio_streams_str is not None:
            audio_streams = json.loads(audio_streams_str, object_hook=lambda d: SimpleNamespace(**d))
        lang_stream = None
        needs_updated = False
        if language:
            test_lang = language.lower()[0:3]
            if audio_streams is not None and hasattr(audio_streams, "streams"):
                for audio_stream in audio_streams.streams:
                    default = audio_stream.disposition.default == 1
                    lang = audio_stream.tags.language
                    if lang == test_lang:
                        lang_stream = audio_stream.index - 1
                        if default is False:
                            needs_updated = True
                    elif default is True:
                        needs_updated = True
                if needs_updated and lang_stream:
                    return lang_stream
    return -1


def update_audio_stream_if_needed(mp4_file: Path, language: Optional[str]) -> bool:
    """
    Returns true if the file had to be edited to have a default audio stream equal to the desired language,
    mostly a concern for apple players (Quicktime/Apple TV/etc.)
    Copies, and potentially updates the default audio stream of a video file.
    """
    random = "".join(choices(population=string.ascii_uppercase + string.digits, k=10))
    work_file = mp4_file.parent / (mp4_file.stem + random + mp4_file.suffix)
    stream = None if language is None else get_audio_stream_for_lang(mp4_file, language)
    if stream is not None and stream >= 0:
        logger.info("Attempt to alter default audio stream of {}", mp4_file)
        with subprocess.Popen(
            [
                "ffmpeg",
                "-i",
                mp4_file,  # input file
                "-map",
                "0",  # copy all stream
                "-disposition:a",
                "none",  # mark all audio streams as not default
                # mark this audio stream as default
                f"-disposition:a:{stream}",
                "default",
                "-c",
                "copy",  # don't re-encode anything.
                work_file,
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        ) as process:
            stdout, stderr = process.communicate()
            success = process.returncode == 0
            if not success:
                logger.info("Could not update audio stream for {}", mp4_file)
                if stderr is not None:
                    logger.info(stderr)
            else:
                logger.warning("Return code: {}", process.returncode)
                mp4_file.unlink()
                shutil.move(work_file, mp4_file)
            return success
    return True


def attempt_fix_corrupt(mp4_file: Path) -> bool:
    """
    Attempt to fix corrupt mp4 files.
    """
    random = "".join(choices(population=string.ascii_uppercase + string.digits, k=10))
    work_file = mp4_file.parent / (mp4_file.stem + random + mp4_file.suffix)
    logger.info("Attempt to fix damaged mp4 file: {}", mp4_file)
    with subprocess.Popen(
        [
            "ffmpeg",
            "-i",
            mp4_file,  # input file
            "-c",
            "copy",  # don't re-encode anything.
            work_file,
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    ) as process:
        stdout, stderr = process.communicate()
        success = process.returncode == 0
        if not success:
            logger.info("Could not fix mp4 files {}", mp4_file)
            if stderr is not None:
                logger.info(stderr)
        else:
            logger.warning("Return code: {}", process.returncode)
            mp4_file.unlink()
            shutil.move(work_file, mp4_file)
        return success
